/* GTK - The GIMP Toolkit
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Choe Hwanjin <krisna@kldp.org>
 */

#ifndef __GTK_IM_CONTEXT_HANGUL_H__
#define __GTK_IM_CONTEXT_HANGUL_H__

#include <gtk/gtkimcontext.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


extern GType gtk_type_im_context_hangul;

#define GTK_TYPE_IM_CONTEXT_HANGUL	gtk_type_im_context_hangul
#define GTK_IM_CONTEXT_HANGUL(obj)	(GTK_CHECK_CAST ((obj), GTK_TYPE_IM_CONTEXT_HANGUL, GtkIMContextHangul))

typedef struct _GtkIMContextHangul	GtkIMContextHangul;
typedef struct _GtkIMContextHangulClass	GtkIMContextHangulClass;
typedef gboolean (*GtkIMContextHangulAutomata) (GtkIMContextHangul *,
					        GdkEventKey *);

struct _GtkIMContextHangul
{
  GtkIMContext object;

  int state;
  gunichar stack[10];
  int index;			/* stack index */
  gunichar choseong;
  gunichar jungseong;
  gunichar jongseong;
  GtkIMContextHangulAutomata automata;
};

struct _GtkIMContextHangulClass
{
  GtkIMContextClass parent_class;
};

void gtk_im_context_hangul_register_type (GTypeModule *type_module,
					  const gchar *type_name);
GtkIMContext *gtk_im_context_hangul_new(void);
void gtk_im_context_hangul_set_automata(GtkIMContextHangul *context_hangul,
					GtkIMContextHangulAutomata automata);

/* asistant function for hangul automata */
#define gtk_im_context_hangul_is_modifier(key)	((key)->state == GDK_CONTROL_MASK || (key)->state == GDK_MOD1_MASK)
#define gtk_im_context_hangul_is_choseong(c)	((c) >=0x1100 && (c) <=0x1112)
#define gtk_im_context_hangul_is_jungseong(c)	((c) >=0x1161 && (c) <=0x1175)
#define gtk_im_context_hangul_is_jongseong(c)	((c) >=0x11a7 && (c) <=0x11c2)
gboolean gtk_im_context_hangul_is_trigger(GdkEventKey *key);
gboolean gtk_im_context_hangul_is_backspace(GdkEventKey *key);
gunichar gtk_im_context_hangul_compchoseong_to_single(gunichar ch);
gunichar gtk_im_context_hangul_compjungseong_to_single(gunichar ch);
gunichar gtk_im_context_hangul_compjongseong_to_single(gunichar ch);
void gtk_im_context_hangul_push(GtkIMContextHangul *context_hangul, gunichar ch);
gunichar gtk_im_context_hangul_pop(GtkIMContextHangul *context_hangul);
gunichar gtk_im_context_hangul_peek(GtkIMContextHangul *context_hangul);

/* commit function */
void gtk_im_context_hangul_commit(GtkIMContextHangul *context_hangul);
void gtk_im_context_hangul_commit_choseong(GtkIMContextHangul *context_hangul,
					   gunichar choseong);
void gtk_im_context_hangul_commit_jungseong(GtkIMContextHangul *context_hangul,
					    gunichar jungseong);
void gtk_im_context_hangul_commit_jongseong(GtkIMContextHangul *context_hangul,
					    gunichar jongseong);
void gtk_im_context_hangul_commit_unicode(GtkIMContextHangul *context_hangul,
			     		  gunichar ch);
gboolean gtk_im_context_hangul_process_nonhangul(GtkIMContextHangul *context_hangul,
						 GdkEventKey *key);
/*
 * global variables for hangul immodule
 */
#define STATE_DIRECT -1 
#define STATE_HANGUL 0 


#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GTK_IM_CONTEXT_HANGUL_H__ */

/* vim: nocindent
 */
